/*====================================+=====================================+
! File CApp.cpp                       ! Copyright (C) 2018-2020 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of kimfo.                                               !
! It is free software: you can redistribute it and/or modify it under the   !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <memory>
#include <ctime>
#include <wx/stdpaths.h>
#include <wx/fileconf.h>
#include <curl/curl.h>
/*-------------------------------------------------------------------------*/
#include "app_info.h"
#include "CApp.h"
#include "CFrame.h"
#include "CDlgPrefs.h"
#include "kimsufi_16x16.xpm"
#include "sys_essential_16x16.xpm"
#include "sys_game_16x16.xpm"
#include "common/sc_download.h"
#include "common/sc_version.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
CApp::~CApp()
{
   /*----------------------------------------------------------------------*/
   delete m_sins_checker ;
   delete m_sins_server ;
   /*----------------------------------------------------------------------*/
   curl_global_cleanup() ;
   /*----------------------------------------------------------------------*/
   if( !m_s_program_to_start_on_exit.empty() )
   {  wxExecute( m_s_program_to_start_on_exit ) ; }
   /*----------------------------------------------------------------------*/
}

/*--( Return 0 if the program has not to exit )----------------------------*/
int CApp::actions_for_install_process()
{
   /*----------------------------------------------------------------------*/
   wxLogNull no_log ; // Suppress useless warnings

   /*-----------------------------------------------------------------------+
   ! The "autoinstall" process has to replace the "kimfo" exec file,        !
   ! the one currently being executed. The problem is that an executed      !
   ! program file is "locked".                                              !
   ! This modification is done in steps.                                    !
   !                                                                        !
   ! The current state of the installation is indicated by the executable   !
   ! base name :                                                            !
   ! - kimfo : Standard execution                                           !
   !   Its tasks are : - delete "kimfo_new" if present                      !
   !                   - Assume expected other tasks :)))                   !
   !                                                                        !
   ! - kimfo_new : After an automatic installation the new executable       !
   !   has been started by "kimfo"                                          !
   !   Its tasks are : - copy itself to the std "kimfo" executable          !
   !                   - start the standard "kimfo" executable              !
   +-----------------------------------------------------------------------*/
   if( m_fn_exe.GetName().EndsWith( "_new" ) )
   {
      /*-------------------------------------------------------------------*/
      wxFileName fn_std( m_fn_exe ) ;
      fn_std.SetName( fn_std.GetName().BeforeFirst( '_' ) ) ;
      /*--( Whatever the result is, a restart is done on std executable )--*/
      set_program_to_start_on_exit( fn_std.GetFullPath() ) ;
      /*--------------------------------------------------------------------+
      ! The standard exe can still been locked.                             !
      ! This loop should not be infinite or there is a bigger problem       !
      +--------------------------------------------------------------------*/
      while( !wxCopyFile( m_fn_exe.GetFullPath(), fn_std.GetFullPath() ) )
      {  wxMilliSleep( 1000 ) ; }
      /*-------------------------------------------------------------------*/
      return 1 ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Remove "new" if present and continue )----------------------------*/
   wxFileName fn_new( m_fn_exe ) ;
   fn_new.SetName( fn_new.GetName() + "_new" ) ;
   /*-----------------------------------------------------------------------+
   ! The new exe can still been locked.                                     !
   ! This loop should not be infinite or there is a bigger problem          !
   +-----------------------------------------------------------------------*/
   while( fn_new.FileExists() && !wxRemoveFile( fn_new.GetFullPath() ) )
   {  wxMilliSleep( 1000 ) ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CApp::OnInit()
{
   /*----------------------------------------------------------------------*/
   if( single_instance() != 0 ) { return false ; }
   /*----------------------------------------------------------------------*/
   curl_global_init( CURL_GLOBAL_ALL ) ;
   /*----------------------------------------------------------------------*/
   m_fn_exe = wxStandardPaths::Get().GetExecutablePath() ;
   /*--( Quit if the install process needs it (cf function header) )-------*/
   if( actions_for_install_process() != 0 ) { return false ; }
   /*----------------------------------------------------------------------*/
   m_locale.Init( wxLANGUAGE_DEFAULT, wxLOCALE_DONT_LOAD_DEFAULT ) ;
   /*--( During development locale dir is located upper ... )--------------*/
   m_locale.AddCatalogLookupPathPrefix(   get_exe_path()
                                        #ifdef __WXDEBUG__
                                        + "/../../.."
                                        #endif // __WXDEBUG__
                                        + "/locale"
                                      ) ;
   /*--( The command line analyse and parsing is done via this call )------*/
   if( !wxApp::OnInit() ) { return false ; }
   /*--( By default, the application name is set to argv[0] )--------------*/
   SetAppName( APP_NAME ) ;
   /*--( Image list )------------------------------------------------------*/
   wxImage::AddHandler( new wxPNGHandler ) ;

   /*--( Init in two steps to set "m_frame" and use it during init )-------*/
   m_frame = new CFrame() ;
   m_frame->init() ;
   /*----------------------------------------------------------------------*/
   return true ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::OnInitCmdLine( wxCmdLineParser &parser )
{
   /*----------------------------------------------------------------------*/
   static const wxCmdLineEntryDesc st_co_a_cmd_line_desc[] =
   {
      { wxCMD_LINE_SWITCH, "h", "help",
        "show this help message",
        wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP
      },
      { wxCMD_LINE_SWITCH, "v", "version",
        "print version information"
      },
      { wxCMD_LINE_OPTION, "l", "language",
        "language (en, en_US, fr_FR ...)"
      },
      { wxCMD_LINE_NONE }
   } ;

   /*----------------------------------------------------------------------*/
   parser.SetLogo( APP_DISPLAY_NAME_ASCII " " APP_VERSION "\n"
                   APP_COPYRIGHT_ASCII " (" APP_HOST ")\n"
                   APP_LICENSE_ASCII "\n"
                 ) ;
   parser.SetDesc( st_co_a_cmd_line_desc ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CApp::OnCmdLineParsed( wxCmdLineParser &parser )
{
   /*--( Version information => quit )-------------------------------------*/
   if( parser.Found( "v" ) )
   {  wxMessageOutput::Get()->Printf( "%s %s", APP_DISPLAY_NAME, APP_VERSION
                                    ) ;
      return false ;
   }

   /*-----------------------------------------------------------------------+
   ! Under MSW the config file will be in the executable's directory        !
   ! Under GTK (and others) the default is used: ~/.appname                 !
   +-----------------------------------------------------------------------*/
   wxString s_config_filename ;
   /*----------------------------------------------------------------------*/
   #ifdef __WXMSW__
      /*-------------------------------------------------------------------*/
      s_config_filename = APP_NAME ".ini" ;
      /*-------------------------------------------------------------------*/
      wxFileName fn_config_file( s_config_filename ) ;
      if( fn_config_file.MakeAbsolute( get_exe_path() ) )
      {  s_config_filename = fn_config_file.GetFullPath() ; }
      /*-------------------------------------------------------------------*/
   #elif defined( __WXGTK__ )
        s_config_filename
      = wxFileConfig::GetLocalFile( APP_NAME ).GetFullPath() ;
   #else
      #error Unknown platform to: determine the config file full name
   #endif // __WXGTK__
   /*-----------------------------------------------------------------------+
   ! Initialization of the configuration file.                              !
   ! If it doesn't exist, it means that the program has never been started  !
   +-----------------------------------------------------------------------*/
   m_boo_first_start_ever = !wxFileExists( s_config_filename ) ;
   /*--( The current config version of this version is set here )----------*/
   open_config_file( s_config_filename, 2 ) ;
   init_config_info() ;

   /*----------------------------------------------------------------------*/
   wxString s_canonical_name ;
   /*--( Language specification precedence: command line, config, system )-*/
   if( !parser.Found( "l", &s_canonical_name ) )
   {  s_canonical_name = M_s_lang.get() ; }
   /*----------------------------------------------------------------------*/
   init_translations( s_canonical_name ) ;
   /*----------------------------------------------------------------------*/
   return true ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_translations( const wxString &s_canonical_name )
{
   /*--( Empty means default meaning the system language ... )-------------*/
   if( !s_canonical_name.empty() )
   {
      /*--( Getting available translations )-------------------------------*/
      wxArrayString as_canon
              = wxTranslations::Get()->GetAvailableTranslations( APP_NAME ) ;
      /*--( Quitting if unknown one )--------------------------------------*/
      if( as_canon.Index( s_canonical_name ) == wxNOT_FOUND ) { return ; }

      /*--( This will only change the texts, not the locales )-------------*/
      wxTranslations *p_translations = new wxTranslations() ;
      p_translations->SetLanguage( s_canonical_name ) ;
      wxTranslations::Set( p_translations ) ;
      /*-------------------------------------------------------------------*/
   }
   /*--( The standard catalog is added for "Copy/Paste/Delete/Cancel ..." )*/
   m_locale.AddCatalog( "wxstd" ) ;
   /*----------------------------------------------------------------------*/
   m_locale.AddCatalog( APP_NAME ) ;
   /*----------------------------------------------------------------------*/
}

/*--( To be used before the reset of config file )-------------------------*/
void CApp::init_config_info_to_keep()
{
   /*--( The entries we want to keep )-------------------------------------*/
   M_s_lang.init( "/current/lang" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::open_config_file( const wxString &s_file, const int i_version )
{
   /*----------------------------------------------------------------------*/
   wxFileConfig *p_conf    ;
   CConfigInt   pi_version ;

   /*----------------------------------------------------------------------*/
   p_conf = new wxFileConfig( wxEmptyString, wxEmptyString, s_file,
                                wxEmptyString,
                                wxCONFIG_USE_LOCAL_FILE
                              | wxCONFIG_USE_RELATIVE_PATH
                            ) ;
   /*-----------------------------------------------------------------------+
   ! By default the strings written are analysed and the environment        !
   ! variables are replaced.                                                !
   +-----------------------------------------------------------------------*/
   p_conf->SetExpandEnvVars( false ) ;
   wxConfigBase::Set( p_conf ) ;

   /*--( Check configuration data version )--------------------------------*/
   pi_version.init( "/profile/version", -1 ) ;
   /*--( If it contains the current version, all is OK )-------------------*/
   if( pi_version.get() == i_version )
   {  return ; }

   /*--( A previous version ? Try to keep info )---------------------------*/
   if( pi_version.get() != -1 )
   {
      /*--( Make a backup of the file )------------------------------------*/
      if( !wxCopyFile( s_file, s_file + ".bak" ) )
      {  return ; }

      /*--( Keep language and playlists if present before reset )----------*/
      init_config_info_to_keep() ;

      /*--------------------------------------------------------------------+
      ! The config file is here but with a "wrong" version value.           !
      ! It has to be "cleaned".                                             !
      +--------------------------------------------------------------------*/
      wxString s_group ;
      long     l_index ;

      /*--( Reset file content after copy keeping some groups )------------*/
      if( p_conf->GetFirstGroup( s_group, l_index ) )
      {
         /*----------------------------------------------------------------*/
         do
         {  /*--( Delete the groups that have not to be kept )-------------*/
            if( s_group != "playlist" ) { p_conf->DeleteGroup( s_group ) ; }
            /*-------------------------------------------------------------*/
         } while( p_conf->GetNextGroup( s_group, l_index ) ) ;
         /*----------------------------------------------------------------*/
      }

      /*--( The variables have to be written before being "reloaded" )-----*/
      M_s_lang.write_val() ;

      /*-------------------------------------------------------------------*/
      init_translations( M_s_lang.get() ) ;

      /*-------------------------------------------------------------------*/
      wxString s_name = APP_DISPLAY_NAME ;
      wxMessageBox( wxString::Format(
                    _( "The configuration file doesn't correspond to the "
                       "current version of %s.\n"
                       "It has been reinitialized (after a backup copy)."
                      ),
                    s_name
                                    ), s_name

                  ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Write new current version value )---------------------------------*/
   pi_version.set( i_version ) ;
   pi_version.write_val() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_config_info()
{
   /*----------------------------------------------------------------------*/
   init_config_info_to_keep() ;
   /*----------------------------------------------------------------------*/
   M_rec_main.init( "/current/rec_main" ) ;
   M_rec_prefs.init( "/current/rec_prefs" ) ;
   M_rec_progress.init( "/current/rec_progress" ) ;
   M_rec_lastversion.init( "/current/rec_lastversion" ) ;
   M_rec_newversion.init( "/current/rec_newversion" ) ;
   /*----------------------------------------------------------------------*/
   M_boo_main_maxi.init( "/current/main_maxi", false ) ;
   M_i_spl_hor_pos.init( "/current/spl_hor_pos", -1 ) ;
   /*----------------------------------------------------------------------*/
   M_s_region.init( "/current/region", "fr" ) ;
   M_i_chkupd_freq.init( "/current/chkupd_freq", sc::CUF_WEEK ) ;
   M_s_chkupd_last.init( "/current/chkupd_last" ) ;
   M_boo_startup_auto_check_avail.init(
                                   "/current/startup_auto_check_avail", false
                                        ) ;
   M_boo_startup_systray_only.init( "/current/startup_systray_only", false );
   M_i_check_avail_period.init( "/current/check_avail_period", 300 ) ;
   M_s_proxy_server.init( "/current/proxy_server" ) ;
   M_i_serv_list_sel.init( "/current/serv_list_sel" ) ;
   M_s_serv_ref_sel.init( "/current/serv_ref_sel" ) ;
   M_s_serv_list_col_width.init( "/current/serv_list_col_width" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::save_config_info()
{
   /*----------------------------------------------------------------------*/
   M_rec_main.save() ;
   M_rec_prefs.save() ;
   M_rec_progress.save() ;
   M_rec_lastversion.save() ;
   M_rec_newversion.save() ;
   /*----------------------------------------------------------------------*/
   M_boo_main_maxi.save() ;
   M_i_spl_hor_pos.save() ;
   /*----------------------------------------------------------------------*/
   M_s_lang.save() ;
   M_s_region.save() ;
   M_i_chkupd_freq.save() ;
   M_s_chkupd_last.save() ;
   M_boo_startup_auto_check_avail.save() ;
   M_boo_startup_systray_only.save() ;
   M_i_check_avail_period.save() ;
   M_s_proxy_server.save() ;
   /*----------------------------------------------------------------------*/
   M_i_serv_list_sel.save() ;
   M_s_serv_ref_sel.save() ;
   M_s_serv_list_col_width.save() ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Return value :                                                            !
! 0    : this instance is the first one and all is ok                       !
! > 0  : this instance is not the first one and all is ok                   !
! < 0  : a problem occurred                                                 !
+--------------------------------------------------------------------------*/
int CApp::single_instance()
{
   /*----------------------------------------------------------------------*/
   wxLogNull no_log ; // Suppress error messages
   /*----------------------------------------------------------------------*/
   m_sins_checker = new wxSingleInstanceChecker( APP_NAME ) ;
   /*--( We become "server" if we are the first instance )-----------------*/
   if( !m_sins_checker->IsAnotherRunning() )
   {
      /*-------------------------------------------------------------------*/
      m_sins_server = new sc::CSIServer() ;
      if( !m_sins_server->Create( APP_SINS_SERVICE ) )
      {  return -1 ; }
      /*-------------------------------------------------------------------*/
      return 0 ;
      /*-------------------------------------------------------------------*/
   }

   /*--( An instance is already running : we want it to raise )------------*/
   sc::CSIClient sins_client ;
   wxConnectionBase *sins_connection
    = sins_client.MakeConnection( "localhost", APP_SINS_SERVICE, APP_NAME ) ;
   /*----------------------------------------------------------------------*/
   if( sins_connection == nullptr )
   {  wxMessageBox( wxString::Format( _( "The current instance of %s may be "
                                         "too busy to respond.\n"
                                         "Please close any open dialog and "
                                         "retry"
                                       ), APP_NAME
                                    ), APP_DISPLAY_NAME
                  ) ;
      return -2 ;
   }
   /*--( Ask main instance to raise by saying nothing )--------------------*/
   sins_connection->Execute( wxEmptyString ) ;
   sins_connection->Disconnect() ;
   delete sins_connection ;
   /*----------------------------------------------------------------------*/
   return 1 ;
   /*----------------------------------------------------------------------*/
}

/*--( Server source data )-------------------------------------------------*/
str_sty st_a_sty[ STY_NB ]
= {
     { "Kimsufi",       kimsufi_16x16_xpm       },
     { "SYS Essential", sys_essential_16x16_xpm },
     { "SYS Game",      sys_game_16x16_xpm      }
  } ;

/*--( Regions alphabetically sorted on the region name )-------------------*/
static const str_region st_a_region[]
= {
    { "ca", "Canada [EN]",
       {  "ca/en/servers.xml",
          "ca/en/essential-servers",
          "ca/en/game-servers"
       }
    },
    { "qc", "Canada [FR]",
       {  "ca/fr/serveurs.xml",
          "ca/fr/serveurs-essential",
          "ca/fr/serveurs-game"
       }
    },
    { "cz", u8"\u010Cesk\u00E1 Republika",
       {  "cz/servery.xml",
          "cz/zakladni-servery",
          "cz/game-servery"
       }
    },
    { "de", "Deutschland",
       {  "de/server.xml",
          "de/essential-server",
          "de/game-server"
       }
    },
    { "es", u8"Espa\u00F1a",
       {  "es/servidores.xml",
          "es/servidores-essential",
          "es/servidores-game"
       }
    },
    { "fr", "France",
       {  "fr/serveurs.xml",
          "fr/serveurs-essential",
          "fr/serveurs-game"
       }
    },
    { "ie", "Ireland",
       {  "en/servers.xml",
          "ie/serveurs-essential",
          "ie/serveurs-game"
       }
    },
    { "it", "Italia",
       {  "it/server.xml",
          "it/server-essential",
          "it/server-game"
       }
    },
    { "lt", "Lietuva",
       {  "lt/serveriai.xml",
          "lt/serveriai-essential",
          "lt/serveriai-game"
       }
    },
    { "ma", "Maroc",
       {  "ma/serveurs.xml",
          "ma/serveurs-essential",
          "ma/serveurs-game"
       }
    },
    { "nl", "Nederland",
       {  "nl/servers.xml",
          "nl/essential-servers",
          "nl/game-servers"
       }
    },
    { "pl", "Polska",
       {  "pl/serwery.xml",
          "pl/serwery-essential",
          "pl/serwery-game"
       }
    },
    { "pt", "Portugal",
       {  "pt/servidores.xml",
          "pt/servidores-essential",
          "pt/servidores-game"
       }
    },
    { "sn", u8"S\u00e9n\u00e9gal",
       {  "sn/serveurs.xml",
          "sn/serveurs-essential",
          "sn/serveurs-game"
       }
    },
    { "fi", "Suomi",
       {  "fi/palvelimet.xml",
          "fi/essential-palvelimet",
          "fi/pelipalvelimet"
       }
    },
    { "tn", "Tunisie",
       {  "tn/serveurs.xml",
          "tn/serveurs-essential",
          "tn/serveurs-game"
       }
    },
    { "uk", "United Kingdom",
       {  "uk/servers.xml",
          "en/essential-servers",
          "en/game-servers"
       }
    },
    { "us", "United States",
       {  "us/en/servers.xml",
          "us/essential-servers",
          "us/game-servers"
       }
    },
    /*--( World in $ )---------------------------------------------------*/
    { "W$", "World [$]",
       {  "us/en/servers.xml",
          "us/essential-servers",
          "us/game-servers"
       }
    },
    /*--( World in Euros )-----------------------------------------------*/
    { "WE", u8"World [\u20AC]",
       {  "en/servers.xml",
          "ie/essential-servers",
          "ie/game-servers"
       }
    },
    /*--( Sentinel )------------------------------------------------------*/
    { nullptr }
    /*--------------------------------------------------------------------*/
  } ;

/*-------------------------------------------------------------------------*/
const str_region *CApp::get_region( int i_num ) const
{  return st_a_region + i_num ; }

/*-------------------------------------------------------------------------*/
const str_region *CApp::find_region( const wxString &s_region ) const
{
   /*----------------------------------------------------------------------*/
   if( !s_region.empty() )
   {
      /*-------------------------------------------------------------------*/
      const char *p_c_region = s_region.c_str() ;
      /*-------------------------------------------------------------------*/
      for( const str_region *p_region = st_a_region ;
           p_region->m_p_c_code != nullptr ;
           ++p_region
         )
      {  if( strcmp( p_region->m_p_c_code, p_c_region ) == 0 )
         {  return p_region ; }
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return nullptr ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_region()
{
   /*----------------------------------------------------------------------*/
   wxString s_region = M_s_region.get() ;

   /*----------------------------------------------------------------------*/
   if( ( m_p_region = find_region( s_region ) ) != nullptr )
   {  return ; }

   /*--( Default to current system language / region )---------------------*/
   s_region = get_canonical_name().Before( '_' ) ;
   if( ( m_p_region = find_region( s_region ) ) != nullptr )
   {  return ; }

   /*--( Default ... )-----------------------------------------------------*/
   m_p_region = &st_a_region[ 0 ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CApp::get_region_page( e_server_type sty ) const
{
   /*----------------------------------------------------------------------*/
   wxASSERT( sty >= 0 && sty < STY_NB ) ;
   return m_p_region->m_a_p_c_sty_page[ sty ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const str_sty *CApp::get_sty( e_server_type sty ) const
{
   /*----------------------------------------------------------------------*/
   wxASSERT( sty >= 0 && sty < STY_NB ) ;
   return &st_a_sty[ sty ] ;
   /*----------------------------------------------------------------------*/
}

/*--( SI : Single Instance. The server receives an execute command )-------*/
bool sc::CSIConnection::OnExec( const wxString &s_topic,
                                const wxString &s_data
                              )
{
   /*--( Frame should not be null at this stage )--------------------------*/
   CFrame *frame = wxGetApp().m_frame ;

   /*--( An empty topic means : just raise ! )-----------------------------*/
   if( frame != nullptr && s_data.empty() )
   {  /*-------------------------------------------------------------------*/
      frame->log_error( wxString::Format(_( "One instance of %s is enough" ),
                                         APP_DISPLAY_NAME
                                        )
                      ) ;
      frame->force_show( true ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return true ;
   /*----------------------------------------------------------------------*/
}

/*--( The program entry point ! )------------------------------------------*/
IMPLEMENT_APP( CApp )
/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                           End of file CApp.cpp                            !
+==========================================================================*/
