/*====================================+=====================================+
! File sc_base.cpp                    ! Copyright (C) 2018-2020 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This is free software: you can redistribute it and/or modify it under the !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/regex.h>
#include <wx/display.h>
#include <wx/tokenzr.h>
#include <wx/clipbrd.h>
#include "common/sc_base.h"
#include "app_info.h"
#include "CApp.h"
#include "CFrame.h"
/*-------------------------------------------------------------------------*/




/*-------------------------------------------------------------------------*/
int sc::set_clipboard( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   if( !wxTheClipboard->Open() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   wxTheClipboard->SetData( new wxTextDataObject( s ) ) ;
   wxTheClipboard->Close() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int sc::get_clipboard( wxString &s )
{
   /*----------------------------------------------------------------------*/
   s.clear() ;
   /*----------------------------------------------------------------------*/
   if( !wxTheClipboard->Open() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   if( wxTheClipboard->IsSupported( wxDF_TEXT ) )
   {  /*-------------------------------------------------------------------*/
      wxTextDataObject data ;
      wxTheClipboard->GetData( data ) ;
      s = data.GetText() ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   wxTheClipboard->Close() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxBitmap sc::get_menu_bmp( const wxArtID &id )
{
   /*----------------------------------------------------------------------*/
   return wxArtProvider::GetBitmap( id, wxART_MENU, wxSize( 16, 16 ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxBitmap sc::get_toolbar_bmp( const wxArtID &id )
{
   /*----------------------------------------------------------------------*/
   return wxArtProvider::GetBitmap( id, wxART_TOOLBAR, wxSize( 32, 32 ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::insert_item( wxMenu *menu, int i_pos, int i_id,
                             const wxString &s_item,
                             const wxBitmap &bmp,
                             const wxString &s_help,
                             wxItemKind mik,
                             wxMenu *p_submenu
                           )
{
   /*----------------------------------------------------------------------*/
   wxMenuItem *mi
            = new wxMenuItem( menu, i_id, s_item, s_help, mik, p_submenu ) ;
   /*----------------------------------------------------------------------*/
   if( bmp.IsOk() ) { mi->SetBitmap( bmp ) ; }
   /*----------------------------------------------------------------------*/
   menu->Insert( i_pos, mi ) ;
   /*----------------------------------------------------------------------*/
   return mi ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::append_item( wxMenu *menu, int i_id,
                             const wxString &s_item,
                             const wxBitmap &bmp,
                             const wxString &s_help,
                             wxItemKind mik,
                             wxMenu *p_submenu
                           )
{
   /*----------------------------------------------------------------------*/
   wxMenuItem *mi
            = new wxMenuItem( menu, i_id, s_item, s_help, mik, p_submenu ) ;
   /*----------------------------------------------------------------------*/
   if( bmp.IsOk() ) { mi->SetBitmap( bmp ) ; }
   /*----------------------------------------------------------------------*/
   menu->Append( mi ) ;
   /*----------------------------------------------------------------------*/
   return mi ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::append_item( wxMenu *menu, int i_id, const wxArtID &artid )
{
   /*----------------------------------------------------------------------*/
   return sc::append_item( menu, i_id,
                           wxGetStockLabel( i_id,  wxSTOCK_WITH_MNEMONIC
                                                 | wxSTOCK_WITH_ACCELERATOR
                                          ),
                           sc::get_menu_bmp( artid )
                         ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::append_item( wxMenu *dst, wxMenuItem *mi_src )
{
   /*----------------------------------------------------------------------*/
   wxMenu *menu_sub = nullptr ;
   if( mi_src->IsSubMenu() )
   {  menu_sub = new wxMenu ;
      sc::append_menu( menu_sub, mi_src->GetSubMenu() ) ;
   }
   /*----------------------------------------------------------------------*/
     wxMenuItem *mi_dst
   = sc::append_item( dst, mi_src->GetId(), mi_src->GetItemLabel(),
                      mi_src->GetBitmap(), mi_src->GetHelp(),
                      mi_src->GetKind(), menu_sub
                    ) ;
   /*----------------------------------------------------------------------*/
   mi_dst->Enable( mi_src->IsEnabled() ) ;
   /*----------------------------------------------------------------------*/
   if( mi_src->IsCheck() || mi_src->IsRadio() )
   {  mi_dst->Check( mi_src->IsChecked() ) ; }
   /*----------------------------------------------------------------------*/
   return mi_dst ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::append_menu( wxMenu *dst, wxMenu *src )
{
   /*----------------------------------------------------------------------*/
   for( const auto &mi_src : src->GetMenuItems() )
   {  append_item( dst, mi_src ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::append_item( wxMenu *dst, wxMenu *src, int i_id )
{
   /*----------------------------------------------------------------------*/
   wxMenuItem *mi = src->FindItem( i_id ) ;
   if( mi != nullptr )
   {  mi = sc::append_item( dst, mi ) ; }
   /*----------------------------------------------------------------------*/
   return mi ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxMenuItem *sc::append_item( wxMenu *dst, wxMenuBar *src, int i_id )
{
   /*----------------------------------------------------------------------*/
   wxMenuItem *mi = src->FindItem( i_id ) ;
   if( mi != nullptr )
   {  mi = sc::append_item( dst, mi ) ; }
   /*----------------------------------------------------------------------*/
   return mi ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxRect sc::get_screen_area_under( const wxWindow *win )
{
   /*----------------------------------------------------------------------*/
   return wxDisplay( wxDisplay::GetFromWindow( win ) ).GetClientArea() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool sc::is_rec_window_ok( const wxRect &rec )
{
   /*--( Something must be visible )---------------------------------------*/
   return    rec.GetWidth()  >= 16
          && rec.GetHeight() >= 16
          && wxDisplay::GetFromPoint( rec.GetTopLeft() ) != wxNOT_FOUND ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::set_tlw_rect( wxWindow *p_win, const wxRect &rect )
{
   /*--( Check the size and position are valid )---------------------------*/
   if( sc::is_rec_window_ok( rect ) )
   {  /*--( Sets size AND position )---------------------------------------*/
      p_win->SetSize( rect ) ;
      /*-------------------------------------------------------------------*/
   }
   else /*--( Use default )------------------------------------------------*/
   {  p_win->CenterOnParent() ;
      p_win->SetSize( p_win->GetMinSize() ) ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::init_tlw_rect( wxWindow *p_win, const CConfigRect &cfg_rect )
{
   /*----------------------------------------------------------------------*/
   p_win->SetMinSize( p_win->GetEffectiveMinSize() ) ;
   /*----------------------------------------------------------------------*/
   sc::set_tlw_rect( p_win, cfg_rect.get() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::save_tlw_rect( wxWindow *p_win, CConfigRect &cfg_rect )
{
   /*----------------------------------------------------------------------*/
   cfg_rect.set( p_win->GetRect() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString sc::sec2hms( int i_sec )
{
   /*----------------------------------------------------------------------*/
   const int co_i_h = i_sec / 60 / 60 ;
   const int co_i_m = i_sec / 60 % 60 ;
   const int co_i_s = i_sec % 60 ;
   wxString s_res ;
   /*----------------------------------------------------------------------*/
   if( co_i_h > 0 )
   {  s_res = wxString::Format( "%dh", co_i_h ) ; }
   /*----------------------------------------------------------------------*/
   if( co_i_h > 0 || co_i_m > 0 )
   {  s_res += wxString::Format( "%dm", co_i_m ) ; }
   /*----------------------------------------------------------------------*/
   if( co_i_s > 0 || s_res.empty() )
   {  s_res += wxString::Format( "%ds", co_i_s ) ; }
   /*----------------------------------------------------------------------*/
   return s_res ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
static void st_center_image( wxImage &img, const wxSize &size_target )
{
   /*----------------------------------------------------------------------*/
   wxPoint pt( ( size_target.GetWidth()  - img.GetWidth()  ) / 2,
               ( size_target.GetHeight() - img.GetHeight() ) / 2
             ) ;
   /*----------------------------------------------------------------------*/
   img.Resize( size_target, pt ) ;
   /*----------------------------------------------------------------------*/
}

/*--( If image is bigger than target then scale down else center )---------*/
wxBitmap sc::scale_down_or_center( const wxBitmap &bmp,
                                   const wxSize &size_target,
                                   bool boo_high_quality
                                 )
{
   /*----------------------------------------------------------------------*/
   const wxSize co_size_bmp( bmp.GetSize() ) ;

   /*--( Minimal control ... )---------------------------------------------*/
   if(    co_size_bmp.GetWidth() <= 0 || co_size_bmp.GetHeight() <= 0
       || size_target.GetWidth() <= 0 || size_target.GetHeight() <= 0
     )
   {  return bmp ; }

   /*--( Same sizes ? No change has to be done )---------------------------*/
   if( co_size_bmp == size_target ) { return bmp ; }

   /*----------------------------------------------------------------------*/
   wxImage img( bmp.ConvertToImage() ) ;

   /*--( If the image is smaller it is centered in the available space )---*/
   if(    co_size_bmp.GetHeight() <= size_target.GetHeight()
       && co_size_bmp.GetWidth()  <= size_target.GetWidth()
     )
   {  /*-------------------------------------------------------------------*/
      st_center_image( img, size_target ) ;
      return img ;
      /*-------------------------------------------------------------------*/
   }

   /*--( If the image is bigger in width or height : it has to rescaled )--*/
   int   i_new_width  = co_size_bmp.GetWidth()  ;
   int   i_new_height = co_size_bmp.GetHeight() ;
   float fl_ratio = static_cast< float >( i_new_width ) / i_new_height ;

   /*--( Proportions have to be kept )-------------------------------------*/
   if( i_new_width > size_target.GetWidth() )
   {  i_new_width  = size_target.GetWidth() ;
      i_new_height = i_new_width / fl_ratio ;
   }
   /*----------------------------------------------------------------------*/
   if( i_new_height > size_target.GetHeight() )
   {  i_new_height = size_target.GetHeight() ;
      i_new_width  = i_new_height * fl_ratio ;
   }
   /*----------------------------------------------------------------------*/
     wxImageResizeQuality e_quality
   = ( boo_high_quality ? wxIMAGE_QUALITY_HIGH : wxIMAGE_QUALITY_NORMAL ) ;
   /*--( Rescale and recenter image in the available space )---------------*/
   img.Rescale( i_new_width, i_new_height, e_quality ) ;
   st_center_image( img, size_target ) ;
   /*----------------------------------------------------------------------*/
   return img ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::set_font_fix( wxWindow *p_win, bool boo_fix )
{
   /*----------------------------------------------------------------------*/
   wxFont cur_font( p_win->GetFont() ) ;
   /*----------------------------------------------------------------------*/
   wxFont win_font( cur_font.GetPointSize(),
                    boo_fix ? wxFONTFAMILY_TELETYPE : wxFONTFAMILY_DEFAULT,
                    cur_font.GetStyle(), cur_font.GetWeight(),
                    cur_font.GetUnderlined(),
                    wxEmptyString, wxFONTENCODING_DEFAULT
                  ) ;
   /*----------------------------------------------------------------------*/
   p_win->SetFont( win_font ) ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! All parameters after "s_re" are supposed to be "wxString *"               !
! representing destination of the different sub-expression in the regex.    !
! For example : match( "toto", ".(.).(.)", &s1, &s2 ) ;                     !
!                                                                           !
! If a param has to be skipped, pass nullptr as its parameter.              !
! For example : match( "toto", ".(.).(.)", nullptr, &s2 ) ;                 !
!                                                                           !
! By default the "ignore case" is set !                                     !
!                                                                           !
! The return value is : GetMatchCount() or < 0 if no match found            !
+--------------------------------------------------------------------------*/
int sc::match( const wxString &s_data, const wxString &s_re ... )
{
   /*----------------------------------------------------------------------*/
   wxRegEx re( s_re, wxRE_ADVANCED | wxRE_ICASE ) ;
   /*----------------------------------------------------------------------*/
   if( !re.IsValid() ) { return -1 ; }
   if( !re.Matches( s_data ) ) { return -2 ; }

   /*----------------------------------------------------------------------*/
   va_list args ;
   va_start( args, s_re ) ;
   /*--( The first param for the match of the whole expression )-----------*/
   const int co_i_nb_param = re.GetMatchCount() ;
   for( int i_num_param = 1 ; i_num_param < co_i_nb_param ; ++i_num_param )
   {  /*-------------------------------------------------------------------*/
      wxString *p_s = va_arg( args, wxString * ) ;
      if( p_s != nullptr ) { *p_s = re.GetMatch( s_data, i_num_param ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   va_end( args ) ;
   /*----------------------------------------------------------------------*/
   return co_i_nb_param ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxColor sc::alt_bg_color( wxColor color_bg )
{
   /*--( Cf "wxListCtrlBase::EnableAlternateRowColours" wxWidgets code )---*/
   int i_alpha = ( color_bg.GetRGB() > 0x808080 ? 97 : 150 ) ;
   return color_bg.ChangeLightness( i_alpha ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString sc::get_debug_filename()
{
   /*----------------------------------------------------------------------*/
   return wxFileName( wxGetApp().get_exe_path(), APP_NAME "_debug.txt"
                    ).GetFullPath() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::trace( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   wxFile f( sc::get_debug_filename(), wxFile::write_append ) ;
   /*----------------------------------------------------------------------*/
   f.Write( wxDateTime::UNow().Format( "%H:%M:%S.%l : " ) + s + "\n" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::CDlgProgress::CDlgProgress( wxWindow *parent )
                : wxDialog( parent, wxID_ANY, wxEmptyString )
{
   /*--( At least under GTK this dialog's icon is not set. It is forced )--*/
   SetIcon( wxGetApp().m_frame->GetIcon() ) ;
   m_bsz_gen = new wxBoxSizer( wxVERTICAL ) ;
   /*----------------------------------------------------------------------*/
     m_stx_msg
   = new wxStaticText( this, wxID_ANY, wxEmptyString,
                       wxDefaultPosition, wxDefaultSize,
                       wxALIGN_CENTRE | wxST_NO_AUTORESIZE
                     ) ;
   m_bsz_gen->Add( m_stx_msg, 1, wxALL | wxEXPAND, 10 ) ;
   /*----------------------------------------------------------------------*/
   m_ga_progress = new wxGauge( this, wxID_ANY, 100 ) ;
   m_ga_progress->SetValue( 0 ) ;
   m_ga_progress->SetMinSize( wxSize( 500, -1 ) ) ;
   m_bsz_gen->Add( m_ga_progress, 1, wxBOTTOM|wxEXPAND|wxLEFT|wxRIGHT, 10 ) ;
   /*----------------------------------------------------------------------*/
   m_bt_quit = new wxButton( this, wxID_CANCEL ) ;
   m_bsz_gen->Add( m_bt_quit, 0, wxALIGN_CENTER_HORIZONTAL | wxALL, 5 ) ;
   m_bt_quit->SetFocus() ;
   m_bt_quit->SetDefault() ;

   /*----------------------------------------------------------------------*/
   SetSizer( m_bsz_gen ) ;
   Layout() ;
   /*----------------------------------------------------------------------*/
   m_bt_quit->Bind( wxEVT_COMMAND_BUTTON_CLICKED, &CDlgProgress::OnBtCancel,
                    this
                  ) ;
   /*----------------------------------------------------------------------*/
}

/*--( SI : Single Instance. The server receives a connection request )-----*/
wxConnectionBase *sc::CSIServer::OnAcceptConnection( const wxString &s_topic)
{
   /*--( The connection is done from another instance ? )------------------*/
   if( s_topic != APP_NAME ) { return nullptr ; }
   /*----------------------------------------------------------------------*/
   return new sc::CSIConnection() ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Menu :                                                                    !
! The accelerator key modifiers "Shift", "Ctrl" and "Alt" etc. must not be  !
! translated. They are "automatically" translated by the menuing system     !
! if the translations are given.                                            !
+--------------------------------------------------------------------------*/
#define DUMMY_TRANSLATIONS wxTRANSLATE( "ctrl" )  \
                           wxTRANSLATE( "shift" ) \
                           wxTRANSLATE( "alt" ) \
                           wxTRANSLATE( "Space" )
#undef DUMMY_TRANSLATIONS

/*--------------------------------------------------------------------------+
! Not very useful but shows the language name in its own language ...       !
! In all languages "French" should stay "Francais",                         !
! "German" should stay "Deutsch" ...                                        !
+--------------------------------------------------------------------------*/
#define DUMMY_TRANSLATED_LANGUAGE_NAMES \
        wxTRANSLATE( "French" ) \
        wxTRANSLATE( "German" )
#undef DUMMY_TRANSLATED_LANGUAGE_NAMES

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                          End of file sc_base.cpp                          !
+==========================================================================*/
